from .base import ScFileException
from .basic import ScFileBasicError, SourceFileNotFound, UnsupportedSuffix, FileIsEmpty, InvalidSignature
from .convert import FileConvertingError, FileParsingError, FileUnsupportedError
from .mcsa import McsaFileError, McsaCountsLimit, McsaUnsupportedVersion, McsaUnsupportedFlags, McsaUnknownLinkCount
from .ol import OlFileError, OlUnknownFourcc, OlInvalidFormat, OlDXNError, OlDXNSize, OlDXNNotSquare



from dataclasses import dataclass, field
from typing import Dict, List

from scfile.consts import McsaModel, Normalization


def scaled_i16(i: float, scale: float) -> float:
    return (i * scale) / Normalization.XYZ_FACTOR

def scaled_u16(i: float, scale: float) -> float:
    return (i * scale) / Normalization.WEIGHT_FACTOR


@dataclass
class Vector:
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

@dataclass
class Texture:
    u: float = 0.0
    v: float = 0.0
    w: float = 1.0

@dataclass
class Curve:
    u: float = 0.0
    v: float = 0.0

@dataclass
class Normals:
    i: float = 0.0
    j: float = 0.0
    k: float = 0.0

@dataclass
class VertexBone:
    ids: Dict[int, int] = field(default_factory=dict)
    weights: Dict[int, float] = field(default_factory=dict)

@dataclass
class Vertex:
    position: Vector = field(default_factory=Vector)
    texture: Texture = field(default_factory=Texture)
    curve: Curve = field(default_factory=Curve)
    normals: Normals = field(default_factory=Normals)
    bone: VertexBone = field(default_factory=VertexBone)

@dataclass
class Polygon:
    v1: int = 0
    v2: int = 0
    v3: int = 0

@dataclass
class Mesh:
    name: bytes = b"name"
    material: bytes = b"material"
    link_count: int = 0
    vertices: List[Vertex] = field(default_factory=list)
    polygons: List[Polygon] = field(default_factory=list)

    def resize_vertices(self, count: int):
        self.vertices = [Vertex() for _ in range(count)]

    def resize_polygons(self, count: int):
        self.polygons = [Polygon() for _ in range(count)]

@dataclass
class Bone:
    name: bytes = b"bone"
    parent_id: int = McsaModel.ROOT_BONE_ID
    position: Vector = field(default_factory=Vector)
    rotation: Vector = field(default_factory=Vector)

@dataclass
class Skeleton:
    bones: List[Bone] = field(default_factory=list)

@dataclass
class Model:
    meshes: List[Mesh] = field(default_factory=list)
    skeleton: Skeleton = field(default_factory=Skeleton)



from dataclasses import dataclass
from typing import Any, Dict, Generator

from scfile.utils.model import Model, Vertex

from .base import BaseOutputFile, OutputData


@dataclass
class ObjOutputData(OutputData):
    model: Model


class ObjFile(BaseOutputFile[ObjOutputData]):
    def write(self) -> None:
        #self._add_material()
        self._add_geometric_vertices()
        self._add_texture_coordinates()
        #self._add_vertex_normals()
        self._ensure_unique_names()
        self._add_polygonal_faces()

    def _add_material(self) -> None:
        self._write_str(f"mtllib {self.filename}.mtl", "\n\n")

    def _add_geometric_vertices(self) -> None:
        for vertex in self._vertices():
            pos = vertex.position
            self._write_str(f"v {pos.x} {pos.y} {pos.z}\n")
        self._write_str("\n")

    def _add_texture_coordinates(self) -> None:
        for vertex in self._vertices():
            tex = vertex.texture
            self._write_str(f"vt {tex.u} {1.0 - tex.v}\n")
        self._write_str("\n")

    def _add_vertex_normals(self) -> None:
        for vertex in self._vertices():
            norm = vertex.normals
            self._write_str(f"vn {norm.i} {norm.j}\n")

        self._write_str("\n")

    def _ensure_unique_names(self):
        names: Dict[bytes, int] = {}

        # Checking names for uniqueness
        for index, mesh in enumerate(self.data.model.meshes):
            name = mesh.name

            names.setdefault(name, 0)
            names[name] += 1
            count = names[name]

            # If name already taken set name with count
            if count > 1:
                mesh.name = name + b"_" + str(count).encode()

            # If name is empty set "noname" with index
            if len(name) < 1:
                mesh.name = b"noname_" + str(index+1).encode()

    def _add_polygonal_faces(self) -> None:
        offset = 0

        for mesh in self.data.model.meshes:
            name = mesh.name.decode()
            self._write_str(f"o {name}\ng {name}\n\n")

            for polygon in mesh.polygons:
                v1 = offset + polygon.v1
                v2 = offset + polygon.v2
                v3 = offset + polygon.v3
                self._write_str(f"f {v1}/{v1} {v2}/{v2} {v3}/{v3}\n")
            self._write_str("\n")

            # Vertex id in mcsa are local to each mesh.
            offset += len(mesh.vertices)

    def _vertices(self) -> Generator[Vertex, Any, Any]:
        for mesh in self.data.model.meshes:
            for vertex in mesh.vertices:
                yield vertex



from typing import List, Dict

from .converter import Converter


class DxnConverter(Converter):
    def to_rgba(self) -> bytes:
        w, h = self.data.width, self.data.height
        imagedata = bytearray(w * h * 4)
        block_size = 16

        offset = 0

        for y in range(h // 4):
            for x in range(w // 4):
                block = self.data.image[offset:offset + block_size]
                offset += block_size

                channel1 = self.unpack_channel(block[:8]) # red
                channel2 = self.unpack_channel(block[8:]) # green

                for j in range(4):
                    for i in range(4):
                        index = ((y * 4 + j) * w + (x * 4 + i)) * 4
                        imagedata[index:index + 4] = [0xFF, channel1[j][i], channel2[j][i], 0xFF]

        return bytes(imagedata)

    def unpack_channel(self, channel: bytes) -> Dict[int, List[int]]:
        color0, color1, indices = channel[0], channel[1], channel[2:]

        palette = self.create_palette(color0, color1)
        indices = self.calculate_indices(indices)

        unpacked = {}

        for i in range(4):
            unpacked[i] = {}
            for j in range(4):
                k = indices & 0x7
                indices >>= 3
                unpacked[i][j] = palette[k]

        return unpacked

    def create_palette(self, color0: int, color1: int) -> List[int]:
        size = 8
        palette = [0] * size
        palette[:2] = color0, color1

        if color0 < color1:
            size -= 2
            palette[6:] = [0, 0xFF]

        div = size - 1

        for i in range(2, size):
            palette[i] = ((size - i) * color0 + (i - 1) * color1) // div

        return palette

    def calculate_indices(self, indices: bytes) -> int:
        return (
            (indices[0]) |
            (indices[1] << 8) |
            (indices[2] << 16) |
            (indices[3] << 24) |
            (indices[4] << 32) |
            (indices[5] << 40)
        )



from typing import List

import lz4.block

from scfile import exceptions as exc
from scfile.consts import Signature, CUBEMAP_FACES
from scfile.files.output.dds import DdsFile, DdsOutputData
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as Format
from scfile.utils.ol.dxn import DxnConverter

from .base import BaseSourceFile


SUPPORTED_FORMATS = [
    b"DXT1",
    b"DXT3",
    b"DXT5",
    b"RGBA8",
    b"BGRA8",
    b"RGBA32F",
    b"DXN_XY"
]


class OlFile(BaseSourceFile):

    output = DdsFile
    signature = Signature.OL
    order = ByteOrder.BIG

    is_cubemap = False

    def to_dds(self) -> bytes:
        return self.convert()

    @property
    def data(self) -> DdsOutputData:
        return DdsOutputData(
            self.width,
            self.height,
            self.mipmap_count,
            self.linear_size,
            self.fourcc,
            self.is_cubemap,
            self.imagedata
        )

    def parse(self) -> None:
        # Read header
        self.width = self.reader.readbin(Format.U32)
        self.height = self.reader.readbin(Format.U32)
        self.mipmap_count = self.reader.readbin(Format.U32)

        # Read encrypted FourCC (dds pixel format)
        self.fourcc = self.reader.readfourcc()

        if self.fourcc not in SUPPORTED_FORMATS:
            raise exc.OlUnknownFourcc(self.path, self.fourcc.decode())

        # Read lz4 uncompressed and compressed sizes
        self._parse_sizes()

        try:
            self.texture_id = self.reader.readstring()
            self._decompress_imagedata()

        except Exception:
            raise exc.OlInvalidFormat(self.path)

        if self.is_dxn:
            self.fourcc = b"BGRA8"
            self.imagedata = DxnConverter(self.data).to_rgba()

    @property
    def linear_size(self) -> int:
        return self.uncompressed[0]

    @property
    def is_rgba(self):
        return self.fourcc in (b"RGBA8", b"BGRA8", b"RGBA32F")

    @property
    def is_dxn(self):
        return self.fourcc in (b"DXN_XY")

    def _read_sizes(self) -> List[int]:
        return [self.reader.readbin(Format.U32) for _ in range(self.mipmap_count)]

    def _parse_sizes(self) -> None:
        self.uncompressed = self._read_sizes()
        self.compressed = self._read_sizes()

    def _decompress_imagedata(self) -> None:
        imagedata = bytearray()

        for mipmap in range(self.mipmap_count):
            imagedata.extend(
                lz4.block.decompress(
                    self.reader.read(self.compressed[mipmap]),
                    self.uncompressed[mipmap]
                )
            )

        self.imagedata = bytes(imagedata)


class OlCubemapFile(OlFile):

    is_cubemap = True

    @property
    def linear_size(self) -> int:
        return 0

    def _read_sizes(self) -> List[List[int]]:
        return [
            [self.reader.readbin(Format.U32) for _ in range(CUBEMAP_FACES)]
            for _ in range(self.mipmap_count)
        ]

    def _parse_sizes(self) -> None:
        self.uncompressed = self._read_sizes()
        self.compressed = self._read_sizes()

    def _decompress_imagedata(self) -> None:
        imagedata = bytearray()

        for mipmap in range(self.mipmap_count):
            for cubemap in range(CUBEMAP_FACES):
                imagedata.extend(
                    lz4.block.decompress(
                        self.reader.read(self.compressed[mipmap][cubemap]),
                        self.uncompressed[mipmap][cubemap]
                    )
                )

        self.imagedata = bytes(imagedata)



from dataclasses import dataclass

from scfile.consts import Magic
from scfile.utils.dds.structure import DDS, BITMASKS

from .base import BaseOutputFile, OutputData


@dataclass
class DdsOutputData(OutputData):
    width: int
    height: int
    mipmap_count: int
    linear_size: int
    fourcc: bytes
    is_cubemap: bool
    image: bytes


class DdsFile(BaseOutputFile[DdsOutputData]):

    magic = Magic.DDS

    def write(self) -> None:
        self._write(DDS.HEADER.SIZE)
        self._write(self.flags)
        self._write(self.data.height)
        self._write(self.data.width)
        self._write(self.pitch_or_linear_size)
        self._write_null(1) # Depth
        self._write(self.data.mipmap_count)
        self._write_null(11) # Reserved

        self._write(DDS.PF.SIZE)
        if self.compressed:
            self._write(DDS.PF.FLAG.FOURCC)
            self._raw_write(self.data.fourcc) # FourCC
            self._write_null(5)

        else:
            self._write(DDS.PF.RGB_FLAGS)
            self._write_null(1) # FourCC
            self._write(DDS.PF.BIT_COUNT)

            for mask in BITMASKS(self.data.fourcc):
                self._write(mask)

        self._write(DDS.TEXTURE | DDS.MIPMAP | DDS.COMPLEX) # Caps1
        self._write(self.cubemap) # Caps2
        self._write_null(3) # Reserved

        self._raw_write(self.data.image)

    @property
    def compressed(self) -> bool:
        return b"DXT" in self.data.fourcc

    @property
    def flags(self) -> int:
        if self.compressed:
            return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.LINEARSIZE
        return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.PITCH

    @property
    def pitch(self) -> int:
        bytes_per_pixel = 4
        aligned_width = (self.data.width * bytes_per_pixel + 3) & ~3
        return aligned_width

    @property
    def pitch_or_linear_size(self) -> int:
        if self.compressed:
            return self.data.linear_size
        return self.pitch

    @property
    def cubemap(self) -> int:
        if self.data.is_cubemap:
            return DDS.CUBEMAPS
        return 0



from enum import IntEnum, auto
from typing import Dict


class Flag(IntEnum):
    """Flag Index"""

    SKELETON = 0
    UV = auto()
    FLAG_3 = auto()
    FLAG_4 = auto()
    FLAG_5 = auto()
    FLAG_6 = auto()


class McsaFlags:
    def __init__(self, version: float):
        self._version = version
        self._flags: Dict[int, bool] = {}

    @property
    def count(self) -> int:
        return {
            7.0: 4,
            8.0: 5,
            10.0: 6
        }.get(self._version, 0)

    @property
    def named_dict(self):
        return {
            Flag(key).name: value
            for key, value in self._flags.items()
        }

    def __getitem__(self, index: int) -> bool:
        return bool(self._flags.get(index, False))

    def __setitem__(self, index: int, value: int):
        self._flags[index] = bool(value)

    def __str__(self):
        return str(self.named_dict)



from scfile.consts import Signature
from scfile.files.output.png import PngFile, PngOutputData

from .base import BaseSourceFile


class MicFile(BaseSourceFile):

    output = PngFile
    signature = Signature.MIC

    def to_png(self) -> bytes:
        return self.convert()

    @property
    def data(self) -> PngOutputData:
        return PngOutputData(self.imagedata)

    def parse(self) -> None:
        self.imagedata = self.reader.read()



from dataclasses import dataclass

from scfile.consts import Magic

from .base import BaseOutputFile, OutputData


@dataclass
class PngOutputData(OutputData):
    image: bytes


class PngFile(BaseOutputFile[PngOutputData]):

    magic = Magic.PNG

    def write(self) -> None:
        self._raw_write(self.data.image)



import os
from pathlib import Path
from typing import NamedTuple, TypeAlias

from .enums import FileSuffix


PathLike: TypeAlias = str | os.PathLike[str] | Path

class Signature(NamedTuple):
    """
    Little-Endian Source File Signature.
    Integer.
    """
    MIC = 0x43494D89
    OL = 0xFD23950A
    MCSA = 0x4153434D

class Magic(NamedTuple):
    """
    Little-Endian Output File Signature.
    List[Integer].
    """
    PNG = [0x89, 0x50, 0x4E, 0x47]
    DDS = [0x44, 0x44, 0x53, 0x20]

class Normalization(NamedTuple):
    # int range + 1
    I8 = VERTEX_WEIGHT = 0x80
    U8 = BONE_WEIGHT = 0x100
    I16 = XYZ_FACTOR = 0x8000
    U16 = WEIGHT_FACTOR = 0x10000

class OlString(NamedTuple):
    SIZE = 17 # 16 bytes + last 0x00 byte
    XOR = ord("g")
    NULL = ord("G")

class McsaModel(NamedTuple):
    ROOT_BONE_ID = -1
    # I dont know specific limit
    # This is done for case when file was read incorrectly
    # So as not to overflow memory
    COUNT_LIMIT = 0x40000

# Files suffixes that can be converted
SUPPORTED_SUFFIXES = {FileSuffix.MIC, FileSuffix.OL, FileSuffix.MCSA}

# Dds cubemap faces count
# +x, -x, +y, -y, +z, -z
CUBEMAP_FACES = 6



from typing import Dict

from scfile import exceptions as exc
from scfile.consts import McsaModel, Normalization, Signature
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as Format
from scfile.files.output.obj import ObjFile, ObjOutputData
from scfile.utils.mcsa.flags import Flag, McsaFlags
from scfile.utils.model import Bone, Mesh, Model, Vertex, scaled_i16, scaled_u16

from .base import BaseSourceFile


SUPPORTED_VERSIONS = [
    7.0,
    8.0,
    10.0
]


class McsaFile(BaseSourceFile):

    output = ObjFile
    signature = Signature.MCSA
    order = ByteOrder.LITTLE

    def to_obj(self) -> bytes:
        return self.convert()

    @property
    def data(self) -> ObjOutputData:
        return ObjOutputData(self.model)

    def parse(self) -> None:
        self.model = Model()

        self._parse_header()
        self._parse_meshes()

        if self.flags[Flag.SKELETON]:
            self._parse_skeleton()

    def _parse_header(self) -> None:
        self._parse_version()
        self._parse_flags()
        self._parse_scales()

    def _parse_version(self):
        self.version = self.reader.readbin(Format.F32)

        if self.version not in SUPPORTED_VERSIONS:
            raise exc.McsaUnsupportedVersion(self.path, self.version)

    def _parse_flags(self):
        self.flags = McsaFlags(self.version)

        for index in range(self.flags.count):
            self.flags[index] = self.reader.readbin(Format.BOOL)

    def _parse_scales(self):
        self.xyz_scale = 1.0
        self.uv_scale = 1.0

        self.xyz_scale = self.reader.readbin(Format.F32)

        if self.flags[Flag.UV]:
            self.uv_scale = self.reader.readbin(Format.F32)

        # ! unknown
        if self.flags[Flag.FLAG_3] and self.version == 10.0:
            self.unknown_scale = self.reader.readbin(Format.F32)

    def _parse_meshes(self) -> None:
        meshes_count = self.reader.readbin(Format.U32)

        for index in range(meshes_count):
            self._parse_mesh()

    def _parse_mesh(self) -> None:
        self.mesh = Mesh()

        self._parse_name_and_material()

        self.bones: Dict[int, int] = {}

        if self.flags[Flag.SKELETON]:
            self._parse_bone_indexes()

        self._parse_counts()

        self._parse_weight_scale()

        # ! unknown
        if self.version == 10.0:
            for _ in range(6):
                self.reader.readbin(Format.F32)

        self._parse_xyz()

        if self.flags[Flag.UV]:
            self._parse_uv()

        # ! unknown
        self._skip_unknown()

        if self.flags[Flag.SKELETON]:
            self._parse_bones()

        self._skip_flag_6()

        self._parse_polygons()

        self.model.meshes.append(self.mesh)

    def _parse_name_and_material(self):
        # I'm still not sure?
        match self.version:
            case 7.0:
                self.mesh.name = self.reader.readstring()
                self.mesh.material = self.reader.readstring()
            case 8.0 | 10.0:
                self.mesh.material = self.reader.readstring()
                self.mesh.name = self.reader.readstring()

    def _parse_bone_indexes(self):
        self.mesh.link_count = self.reader.readbin(Format.U8)
        total_bones = self.reader.readbin(Format.U8)

        for index in range(total_bones):
            self.bones[index] = self.reader.readbin(Format.I8)

    def _parse_counts(self):
        self.vertices_count = self.reader.mcsa_counts()
        self.polygons_count = self.reader.mcsa_counts()

        self.mesh.resize_vertices(self.vertices_count)
        self.mesh.resize_polygons(self.polygons_count)

    def _parse_weight_scale(self):
        self.weight_scale = 1.0

        # ! unknown
        if self.flags[Flag.UV]:
            self.weight_scale = self.reader.readbin(Format.F32)

    def _parse_xyz(self) -> None:
        vertices = self.reader.mcsa_xyz(self.vertices_count)

        for vertex, (x, y, z, w) in zip(self.mesh.vertices, vertices):
            vertex.position.x = scaled_i16(x, self.xyz_scale)
            vertex.position.y = scaled_i16(y, self.xyz_scale)
            vertex.position.z = scaled_i16(z, self.xyz_scale)
            vertex.position.w = scaled_u16(w, self.weight_scale)

    def _parse_uv(self) -> None:
        vertices = self.reader.mcsa_uv(self.vertices_count)

        for vertex, (u, v) in zip(self.mesh.vertices, vertices):
            vertex.texture.u = scaled_i16(u, self.uv_scale)
            vertex.texture.v = scaled_i16(v, self.uv_scale)

    def _skip_vertices(self) -> None:
        # 4 bytes per vertex
        self.reader.read(self.vertices_count * 4)

    def _skip_unknown(self) -> None:
        if self.flags[Flag.FLAG_3]:
            self._skip_vertices()

        if self.flags[Flag.FLAG_4]:
            self._skip_vertices()

        if self.flags[Flag.FLAG_5]:
            self._skip_vertices()

    def _parse_bones(self) -> None:
        match self.mesh.link_count:
            case 0:
                pass
            case 1 | 2:
                self._parse_bone_packed()
            case 3 | 4:
                self._parse_bone_plains()
            case _:
                raise exc.McsaUnknownLinkCount(self.path, self.mesh.link_count)

    def _parse_bone_packed(self) -> None:
        for vertex in self.mesh.vertices:
            self._parse_bone_id(vertex, 2)
            self._parse_bone_weight(vertex, 2)

    def _parse_bone_plains(self) -> None:
        for vertex in self.mesh.vertices:
            self._parse_bone_id(vertex, 4)

        for vertex in self.mesh.vertices:
            self._parse_bone_weight(vertex, 4)

    def _parse_bone_id(self, vertex: Vertex, size: int) -> None:
        for index in range(size):
            bone_id = self.reader.readbin(Format.I8)
            vertex.bone.ids[index] = self.bones.get(bone_id, McsaModel.ROOT_BONE_ID)

    def _parse_bone_weight(self, vertex: Vertex, size: int) -> None:
        for index in range(size):
            bone_weight = self.reader.readbin(Format.I8)
            vertex.bone.weights[index] = bone_weight / Normalization.BONE_WEIGHT

    def _skip_flag_6(self):
        if self.flags[Flag.FLAG_6]:
            self._skip_vertices()

    def _parse_polygons(self) -> None:
        polygons = self.reader.mcsa_polygons(self.polygons_count)

        # In obj vertex indexes starts with 1, but in mcsa with 0.
        # So we increase each one by one.
        for polygon, (v1, v2, v3) in zip(self.mesh.polygons, polygons):
            polygon.v1 = v1 + 1
            polygon.v2 = v2 + 1
            polygon.v3 = v3 + 1

    def _parse_skeleton(self) -> None:
        # Still no export support yet
        return

        bones_count = self.reader.readbin(Format.I8)

        for index in range(bones_count):
            self._parse_bone(index)

    def _parse_bone(self, index: int) -> None:
        self.bone = Bone()

        self.bone.name = self.reader.readstring()

        parent_id = self.reader.readbin(Format.I8)
        self.bone.parent_id = parent_id if parent_id != index else McsaModel.ROOT_BONE_ID

        self._parse_bone_position()
        self._parse_bone_rotation()

        self.model.skeleton.bones.append(self.bone)

    def _parse_bone_position(self):
        self.bone.position.x = self.reader.readbin(Format.F32)
        self.bone.position.y = self.reader.readbin(Format.F32)
        self.bone.position.z = self.reader.readbin(Format.F32)

    def _parse_bone_rotation(self):
        self.bone.rotation.x = self.reader.readbin(Format.F32)
        self.bone.rotation.y = self.reader.readbin(Format.F32)
        self.bone.rotation.z = self.reader.readbin(Format.F32)



from pathlib import Path
from typing import Optional

import click
from rich import print

from scfile.exceptions import ScFileException
from scfile.utils import convert


def convert_file(src: Path, dest: Optional[Path] = None):
    try:
        convert.auto(src, dest)

    except ScFileException as err:
        print(f"\n[b red]Error:[/] {err}")

    except Exception as err:
        print(f"\n[b red]Unknown Error:[/] '{src}' - {err}")

def convert_files(files: tuple[Path], dest: Optional[Path] = None):
    for src in files:
        convert_file(src, dest)

@click.command(no_args_is_help=True)
@click.argument(
    "files",
    type=click.Path(path_type=Path, exists=True, readable=True, dir_okay=False),
    nargs=-1
)
@click.option(
    "-O", "--output",
    help="Optional path to output. Defaults to source path with new suffix.",
    type=click.Path(path_type=Path, exists=False, writable=True, dir_okay=True),
    multiple=True,
    nargs=1,
)
def main(files: tuple[Path], output: Optional[tuple[Path]] = None):
    if not output:
        convert_files(files)
        return

    first_output = output[0]
    if first_output.is_dir():
        convert_files(files, first_output)
        return

    for src, dest in zip(files, output):
        convert_file(src, dest)


if __name__ == "__main__":
    try:
        main()

    except KeyboardInterrupt:
        print("\n\n[b yellow]Operation aborted.[/]")



from abc import ABC, abstractmethod, abstractproperty
from io import BytesIO
from pathlib import Path
from typing import Type

from scfile import exceptions as exc
from scfile.consts import PathLike
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as Format
from scfile.files.output.base import BaseOutputFile, OutputData
from scfile.reader import BinaryReader


class BaseSourceFile(ABC):
    DEFAULT_VALIDATE = True

    def __init__(self, path: PathLike, validate_signature: bool = DEFAULT_VALIDATE):
        self.path = Path(path)
        self.validate = validate_signature

        self.reader = BinaryReader(self.path, self.order)
        self.buffer = BytesIO()

        self.check_filesize()
        self.check_signature()

    output: Type[BaseOutputFile]
    """Output file model."""

    signature: int
    """Signature of source file. For example: `0x38383431`."""

    order: ByteOrder = BinaryReader.DEFAULT_BYTEORDER
    """Reader bytes order format."""

    def convert(self) -> bytes:
        """Convert source file. Return output file bytes."""
        self.parse()
        self.output(self.path, self.data, self.buffer).create()
        return self.result

    @abstractproperty
    def data(self) -> OutputData:
        """Return output data."""
        ...

    @abstractmethod
    def parse(self) -> None:
        """Parse source file."""
        ...

    @property
    def result(self) -> bytes:
        """Result bytes of conversion."""
        return self.buffer.getvalue()

    @property
    def filename(self) -> str:
        """Filename of source file."""
        return self.path.stem

    @property
    def filesize(self) -> int:
        """Size of source file."""
        return self.path.stat().st_size

    def check_filesize(self) -> None:
        """Check if file size is valid."""
        if self.filesize < len(str(self.signature)):
            raise exc.FileIsEmpty(self.path)

    def validate_signature(self, signature: int) -> bool:
        """Validate signature of source file."""
        return signature == self.signature

    def check_signature(self) -> None:
        """Check if signature of source file is valid."""
        signature = self.reader.readbin(Format.U32, ByteOrder.LITTLE)

        if self.validate and not self.validate_signature(signature):
            raise exc.InvalidSignature(self.path, signature, self.signature)

    def close(self):
        self.reader.close()
        self.buffer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __str__(self):
        return f"<{self.__class__.__name__}> path='{self.path.as_posix()}' "



import struct
from io import FileIO
from pathlib import Path
from typing import Any, Optional

from scfile.consts import McsaModel, Normalization, OlString, PathLike
from scfile.exceptions import McsaCountsLimit

from scfile.enums import ByteOrder
from scfile.enums import StructFormat as Format


class BinaryReader(FileIO):
    DEFAULT_BYTEORDER = ByteOrder.STANDARD

    def __init__(self, path: PathLike, order: ByteOrder = DEFAULT_BYTEORDER):
        super().__init__(path, mode="rb")
        self.path = Path(path)
        self.order = order

    def readbin(self, fmt: str, order: Optional[ByteOrder] = None) -> Any:
        """Read binary value from stream."""

        order = order or self.order
        return self.unpack(f"{order}{fmt}")[0]

    def readstring(self, order: Optional[ByteOrder] = None) -> bytes:
        """Read an MCSA/OL file string from stream."""

        size = self.readbin(Format.U16, order)
        return self.unpack(f"{size}s")[0]

    def readfourcc(self) -> bytes:
        """Read an OL file FourCC string from stream."""

        # read and skip last 0x00 byte
        string = self.read(OlString.SIZE)[:-1]

        # xor byte if it's is not null
        return bytes(byte ^ OlString.XOR for byte in string if byte != OlString.NULL)

    def mcsa_counts(self):
        """Read MCSA counts from stream and validate against COUNT_LIMIT."""

        counts = self.readbin(Format.U32)
        if counts > McsaModel.COUNT_LIMIT:
            raise McsaCountsLimit(self.path, counts)
        return counts

    def mcsa_xyz(self, vertices_count: int):
        """Read MCSA vertices XYZ coordinates from stream."""

        fmt = Format.I16 * 3 + Format.U16

        return self._read_mcsa(fmt, vertices_count)

    def mcsa_uv(self, vertices_count: int):
        """Read MCSA vertices UV coordinates from stream."""

        fmt = Format.I16 * 2

        return self._read_mcsa(fmt, vertices_count)

    def mcsa_polygons(self, polygons_count: int):
        """Read MCSA polygons from stream."""

        # If it works, dont touch it
        u32 = polygons_count * 3 >= Normalization.U16

        fmt = Format.U32 if u32 else Format.U16
        fmt *= 3

        return self._read_mcsa(fmt, polygons_count)

    def _read_mcsa(self, fmt: str, count: int):
        """Read MCSA data based on given format and count."""

        data = self._unpack_mcsa(fmt, count)
        return self._split_mcsa(fmt, data)

    def _unpack_mcsa(self, fmt: str, count: int):
        """Unpack MCSA data from stream."""

        fmt *= count
        return self.unpack(fmt)

    def _split_mcsa(self, fmt: str, data: Any):
        """Split MCSA data into an array of arrays."""

        size = len(fmt)
        return [
            data[i:i+size]
            for i in range(0, len(data), size)
        ]

    def unpack(self, fmt: str) -> Any:
        """Unpack binary data based on the given format."""

        size = struct.calcsize(str(fmt))
        data = self.read(size)
        return struct.unpack(fmt, data)



from .files import McsaFile, MicFile, OlFile
from .utils import convert
from .utils.convert import mcsa_to_obj, mic_to_png, ol_to_dds

__all__ = (
    "McsaFile", "MicFile", "OlFile",
    "mcsa_to_obj", "mic_to_png", "ol_to_dds",
    "convert"
)



from pathlib import Path

from scfile.consts import PathLike

from .convert import FileConvertingError, FileParsingError, FileUnsupportedError


class OlFileError(FileConvertingError):
    """Base ol texture file exception."""

    def __str__(self):
        return f"Texture '{self.posix_path}'"

class OlInvalidFormat(OlFileError, FileParsingError):
    """Exception occurring when cannot uncompress image."""

    def __str__(self):
        return f"{super().__str__()} has invalid format."

class OlUnknownFourcc(OlFileError, FileUnsupportedError):
    """Exception occurring when texture have unknown fourcc."""

    def __init__(self, path: PathLike, fourcc: str):
        self.path = Path(path)
        self.fourcc = fourcc

    def __str__(self):
        return f"{super().__str__()} has unknown format '{self.fourcc}'"

class OlDXNError(OlFileError, FileParsingError):
    """Exception occurring due DXN XY unpacking."""
    pass

class OlDXNSize(OlDXNError):
    pass

class OlDXNNotSquare(OlDXNError):
    pass



from pathlib import Path

from scfile.consts import PathLike

from .base import ScFileException


class FileConvertingError(ScFileException):
    """Exception occurring when file converting."""

    def __init__(self, path: PathLike):
        self.path = Path(path)

    @property
    def posix_path(self) -> str:
        return self.path.as_posix()

class FileParsingError(ScFileException):
    """Exception occurring due incorrect file parsing. (Or file incorrect itself)."""
    pass

class FileUnsupportedError(ScFileException):
    """Exception occurring intentionally due incorrect file parsing."""
    pass



from pathlib import Path
from typing import Optional, Type

from scfile import exceptions as exc
from scfile.consts import PathLike
from scfile.enums import FileSuffix
from scfile.files import McsaFile, MicFile, OlCubemapFile, OlFile
from scfile.files.source.base import BaseSourceFile


def mcsa_to_obj(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.mcsa` file to `.obj`.

    Args:
        source: Full path to `.mcsa` file.
        output (optional): Full path to output `.obj` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_obj("C:/file.mcsa", "C:/file.obj")`
    """

    _convert(source, output, McsaFile, FileSuffix.OBJ)


def mic_to_png(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.mic` file to `.png`.

    Args:
        source: Full path to `.mic` file.
        output (optional): Full path to output `.png` file.
        Defaults to source path with new suffix.

    Example:
        `mic_to_png("C:/file.mic", "C:/file.png")`
    """

    _convert(source, output, MicFile, FileSuffix.PNG)


def ol_to_dds(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting `.ol` file to `.dds`.

    Args:
        source: Full path to `.ol` file.
        output (optional): Full path to output `.dds` file.
        Defaults to source path with new suffix.

    Example:
        `ol_to_dds("C:/file.ol", "C:/file.dds")`
    """

    try:
        _convert(source, output, OlFile, FileSuffix.DDS)

    except exc.OlInvalidFormat:
        _convert(source, output, OlCubemapFile, FileSuffix.DDS)


def auto(source: PathLike, output: Optional[PathLike] = None):
    """
    Automatically determines which format convert to.

    Args:
        source: Full path to encrypted file.
        output (optional): Full path to output decrypted file.
        Defaults to source path with new suffix.

    Raises:
        UnsupportedSuffix - if source suffix not in `.mic`, `.ol`, `.mcsa`.

    Example:
        `auto("C:/file.mic", "C:/file.png")`
    """

    path = Path(source)

    match path.suffix.lstrip("."):
        case FileSuffix.MIC:
            mic_to_png(source, output)

        case FileSuffix.OL:
            ol_to_dds(source, output)

        case FileSuffix.MCSA:
            mcsa_to_obj(source, output)

        case _:
            raise exc.UnsupportedSuffix(path)


def _convert(
    source: PathLike,
    output: Optional[PathLike],
    converter: Type[BaseSourceFile],
    new_suffix: str
):
    src = Path(source)
    new_src = src.with_suffix(f".{new_suffix}")

    dest = Path(output) if output else new_src

    # Check that file exists
    if not src.exists() or not src.is_file():
        raise exc.SourceFileNotFound(src)

    # If destination is directory save file in this folder
    if dest.is_dir():
        dest = Path(dest, new_src.name)

    # Get converted bytes using context manager
    with converter(src) as encrypted:
        converted = encrypted.convert()

    # Save converted bytes in destination file
    with open(dest, "wb") as fp:
        fp.write(converted)



from pathlib import Path

from scfile.consts import PathLike

from .base import ScFileException


class ScFileBasicError(ScFileException):
    """Basic files exception. Occurring when file processing."""

    def __init__(self, path: PathLike):
        self.path = Path(path)

    @property
    def posix_path(self) -> str:
        return self.path.as_posix()

    def __str__(self):
        return f"File '{self.posix_path}'"

class SourceFileNotFound(ScFileBasicError):
    """Exception occurring when source (encrypted) file not found or does not exists."""

    def __str__(self):
        return f"{super().__str__()} not found (not exists)."

class UnsupportedSuffix(ScFileBasicError):
    """Exception occurring when file have unknown suffix."""

    def __str__(self):
        return f"{super().__str__()} with suffix '{self.path.suffix}' is unsupported."

class FileIsEmpty(ScFileBasicError):
    """Exception occurring when file is empty."""

    def __str__(self):
        return f"{super().__str__()} is empty."

class InvalidSignature(ScFileBasicError):
    """Exception occurring when file signature does not match file type."""

    def __init__(self, path: PathLike, signature: int, valid_signature: int):
        self.path = Path(path)
        self.signature = signature
        self.valid_signature = valid_signature

    def __str__(self):
        return (
            f"{super().__str__()} has invalid signature - "
            f"{hex(self.signature)} != {hex(self.valid_signature)}. "
            "(File suffix does not match file type)."
        )



from pathlib import Path

from scfile.consts import PathLike
from scfile.utils.mcsa.flags import McsaFlags

from .convert import FileConvertingError, FileParsingError, FileUnsupportedError


class McsaFileError(FileConvertingError):
    """Base mcsa model file exception."""

    def __str__(self):
        return f"Model '{self.posix_path}'"

class McsaCountsLimit(McsaFileError, FileParsingError):
    """Exception occurring when vertices or polygons count is not read correctly."""

    def __init__(self, path: PathLike, counts: int):
        self.path = Path(path)
        self.counts = counts

    def __str__(self):
        return f"{super().__str__()} has invalid format."

class McsaUnknownLinkCount(McsaFileError, FileParsingError):
    """Exception occurring when model skeleton bones have unknown link count."""

    def __init__(self, path: PathLike, link_count: int):
        self.path = Path(path)
        self.link_count = link_count

    def __str__(self):
        return f"{super().__str__()} has unknown bones link count: {self.link_count}"

class McsaUnsupportedVersion(McsaFileError, FileUnsupportedError):
    """Exception occurring when model version unsupported."""

    def __init__(self, path: PathLike, version: float):
        self.path = Path(path)
        self.version = version

    def __str__(self):
        return f"{super().__str__()} has unsupported version: {self.version}"

class McsaUnsupportedFlags(McsaFileError, FileUnsupportedError):
    """Exception occurring when model flags unsupported."""

    def __init__(self, path: PathLike, flags: McsaFlags):
        self.path = Path(path)
        self.flags = flags

    def __str__(self):
        return f"{super().__str__()} has unsupported flags: {self.flags}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scfile',
 'scfile.exceptions',
 'scfile.files',
 'scfile.files.output',
 'scfile.files.source',
 'scfile.reader',
 'scfile.utils',
 'scfile.utils.dds',
 'scfile.utils.mcsa',
 'scfile.utils.ol']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.7,<9.0.0', 'lz4>=4.3.2,<5.0.0']

entry_points = \
{'console_scripts': ['build = build:build']}

setup_kwargs = {
    'name': 'sc-file',
    'version': '2.0.0',
    'description': '',
    'long_description': 'SC FILES\n==========================\n\nLibrary and Utility for converting encrypted ``stalcraft`` game files, such as models and textures into well-known formats.\n\nYou can use compiled cli utility from `Releases <https://github.com/onejeuu/sc-file/releases>`_ page.\n\n\n⚠️ Disclaimer\n-------------\n\n**Do not use game assets folder directly.**\n\nYou can get banned for any changes in game client.\n\nCopy files you need to another folder and work there.\n\n\n📁 Formats\n----------\n\n.. list-table::\n   :widths: 20 20 20\n\n   * - Type\n     - Source format\n     - Output format\n   * - Model\n     - .mcsa\n     - .obj\n   * - Texture\n     - .mic\n     - .png\n   * - Texture\n     - .ol\n     - .dds\n\n\n💻 CLI Utility\n--------------\n\nUsage\n~~~~~\n\nYou can drag and drop one or multiple files to ``scfile.exe``.\n\nFrom bash:\n\n.. code:: bash\n\n    scfile [OPTIONS] [FILES]...\n\nArguments\n~~~~~~~~~\n\n- ``FILES``: **List of file paths to be converted**. Multiple files should be separated by **spaces**. Accepts both full and relative paths. **Does not accept directory**.\n\nOptions\n~~~~~~~\n\n- ``-O``, ``--output``: **One path to output file or directory**. Can be specified multiple times for different output files or directories. If not specified, file will be saved in the same directory with a new suffix. You can specify multiple ``FILES`` and a single ``--output`` directory.\n\nExamples\n~~~~~~~~\n\n1. Convert a single file:\n\n    .. code:: bash\n\n        scfile file.mcsa\n\n    Will be saved in the same directory with a new suffix.\n\n2. Convert a single file with a specified path or name:\n\n    .. code:: bash\n\n        scfile file.mcsa --output path/to/file.obj\n\n3. Convert multiple files to a specified directory:\n\n    .. code:: bash\n\n        scfile file1.mcsa file2.mcsa --output path/to/folder\n\n4. Convert multiple files with explicitly specified output files:\n\n    .. code:: bash\n\n        scfile file1.mcsa file2.mcsa -O 1.obj -O 2.obj\n\n    If the count of ``FILES`` and ``-O`` is different, as many files as possible will be converted.\n\n5. Convert all ``.mcsa`` files in the current directory:\n\n    .. code:: bash\n\n        scfile *.mcsa\n\n    In this case, ``-O`` accepts only a directory. Subfolders are not included.\n\n6. Convert all ``.mcsa`` files with subfolders to a specified directory:\n\n    .. code:: bash\n\n        scfile **/*.mcsa -O path/to/folder\n\n    In this case, ``-O`` accepts only a directory. With ``-O`` specified, the folder structure is not duplicated.\n\n\n📚 Library\n----------\n\nInstall\n~~~~~~~\n\nPip\n~~~\n\n.. code:: bash\n\n    pip install sc-file -U\n\nManual\n~~~~~~\n\n.. code:: bash\n\n    git clone git@github.com:onejeuu/sc-file.git\n\n.. code:: bash\n\n    cd sc-file\n\n.. code:: bash\n\n    poetry install\n\nUsage\n~~~~~\n\nSimple\n^^^^^^\n\n.. code:: python\n\n    from scfile import convert\n\n    # Output path is optional.\n    # Defaults to source path with new suffix.\n    convert.mcsa_to_obj("path/to/file.mcsa", "path/to/file.obj")\n    convert.mic_to_png("path/to/file.mic", "path/to/file.png")\n    convert.ol_to_dds("path/to/file.ol", "path/to/file.dds")\n\nAdvanced\n^^^^^^^^\n\n.. code:: python\n\n    from scfile import McsaFile\n\n    with McsaFile("path/to/file.mcsa") as mcsa:\n        obj: bytes = mcsa.to_obj()\n\n    with open("path/to/file.obj", "wb") as fp:\n        fp.write(obj)\n\nBuild\n~~~~~\n\n.. code:: bash\n\n    poetry install\n\n.. code:: bash\n\n    poetry run build\n',
    'author': 'onejeuu',
    'author_email': 'bloodtrail@beber1k.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<3.13',
}


setup(**setup_kwargs)



import struct
from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import Path
from typing import Any, Generic, List, Optional, TypeVar

from scfile.consts import PathLike
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as Format


class OutputData(ABC):
    """Can contain any type of data."""
    pass


T = TypeVar("T", bound="OutputData")


class BaseOutputFile(ABC, Generic[T]):
    DEFAULT_BYTEORDER = ByteOrder.LITTLE
    DEFAULT_FORMAT = Format.U32

    def __init__(
        self,
        path: PathLike,
        data: T,
        buffer: BytesIO,
    ):
        self.path = Path(path)
        self.data = data
        self.buffer = buffer

    magic: Optional[List[int]] = None
    """Magic value (signature) of output file. For example: `[0x89, 0x50, 0x4E, 0x47]`."""

    def create(self) -> bytes:
        """Create output file. Return output file bytes."""
        self.add_magic()
        self.write()
        return self.result

    @property
    def result(self) -> bytes:
        """Result bytes of conversion."""
        return self.buffer.getvalue()

    @abstractmethod
    def write(self) -> None:
        """Writes output file in buffer."""
        ...

    @property
    def filename(self) -> str:
        """Filename of output file."""
        return self.path.stem

    def add_magic(self):
        if self.magic:
            self._raw_write(bytes(self.magic))

    def _raw_write(self, data: bytes) -> None:
        self.buffer.write(data)

    def _write_null(self, count: int):
        self._raw_write(b'\x00' * 4 * count)

    def _write_str(self, *data: Any) -> None:
        string = "".join([str(i) for i in data])
        self.buffer.write(string.encode())

    def _write(
        self,
        v: Any,
        fmt: Format = DEFAULT_FORMAT,
        order: ByteOrder = DEFAULT_BYTEORDER
    ) -> None:
        data = struct.pack(f"{order}{fmt}", v)
        self.buffer.write(data)

    def __str__(self):
        return f"<{self.__class__.__name__}> frompath='{self.path.as_posix()}'"
